from typing import Union

from torch.nn import functional
from torchmetrics import Metric
import torchmetrics


def suggest_string(candidates, lookfor, typ):
    print(lookfor, f'is not a valid choice for {typ}. Possible choices are:', candidates)


def get_loss_fun(loss: Union[str, callable]):
    if callable(loss):
        return loss
    try:
        return getattr(functional, loss)
    except AttributeError:
        raise ValueError(suggest_string(dir(functional), loss, typ='loss'))


def get_metric_fun(metric: Union[str, Metric]):
    if isinstance(metric, Metric):
        return metric
    try:
        return getattr(torchmetrics, metric)()
    except AttributeError:
        raise ValueError(suggest_string(dir(torchmetrics), metric, typ='metric'))



from typing import Union

from attr import define
from sklearn.pipeline import Pipeline, FeatureUnion
from torch.utils.data import Dataset


class FTDataset(Dataset):
    def __init__(self, items,
                 get_x: Union[callable, Pipeline, FeatureUnion],
                 get_y: Union[callable, Pipeline, FeatureUnion]):
        self.items = items
        self.get_x = get_x
        self.get_y = get_y

    def __len__(self):
        return len(self.items)

    def __getitem__(self, idx):
        item = self.items[idx]

        if isinstance(self.get_x, (Pipeline, FeatureUnion)):
            x = self.get_x.transform([item])
        else:
            x = self.get_x(item)
        if isinstance(self.get_y, (Pipeline, FeatureUnion)):
            y = self.get_y.transform([item])
        else:
            y = self.get_y(item)

        return x, y


@define
class DatasetBlueprint:
    get_x: Union[callable, Pipeline, FeatureUnion]
    get_y: Union[callable, Pipeline, FeatureUnion]

    def __call__(self, items):
        return FTDataset(items, self.get_x, self.get_y)



from sklearn import pipeline, preprocessing
import numpy as np


class CallableFunctionTransformer(pipeline.FunctionTransformer):
    def __call__(self, X):
        return super(CallableFunctionTransformer, self).transform(X)


def make_transformer(f, *args, **kwargs):
    return CallableFunctionTransformer(f, *args, **kwargs)


@make_transformer
def read_tifffile(filename: str):
    try:
        from tifffile import imread
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f'You must install tifffile (pip install tifffile) for this to work.')
    return imread(filename)


def _scale(array):
    shape = array.shape
    array = preprocessing.minmax_scale(array.reshape(-1, 1), feature_range=(0, 255)).astype(np.uint8)
    return array.reshape(*shape)


@make_transformer
def tiff_to_rgb(image):
    rgb = image[:, :, 2], image[:, :, 1], image[:, :, 0]
    rgb = [_scale(i) for i in rgb]
    return np.stack(rgb, axis=-1)

